/*hdr
**	FILE:		8bit.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		20/3/97
**	DESCRIPTION:	8 bit primitives
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"

// Dln 24->8 bit quantization library
#ifdef DLN
extern "C" {
#include "dl1quant.h"
}
#endif

/****************************** Classes *************************************************************************************/
bool GdcApp8::SetSurface(GdcSurface *d, GdcPalette *p, GdcSurface *a)
{
	if (d AND d->Bits == 8)
	{
		Dest = d;
		Pal = p;
		Ptr = d->Base;
		Alpha = 0;
		return TRUE;
	}
	return FALSE;
}

void GdcApp8::SetPtr(int x, int y)
{
	ASSERT(Dest AND Dest->Base);
	Ptr = Dest->Base + ((y * Dest->Line) + x);
}

void GdcApp8::IncX()
{
	Ptr++;
}

void GdcApp8::IncY()
{
	Ptr += Dest->Line;
}

void GdcApp8::IncPtr(int X, int Y)
{
	Ptr += (Y * Dest->Line) + X;
}

COLOUR GdcApp8::Get()
{
	return *Ptr;
}

// 8 bit set sub functions
void GdcApp8Set::Set()
{
	*Ptr = c;
}

void GdcApp8Set::VLine(int height)
{
	while (height--)
	{
		*Ptr = c;
		Ptr += Dest->Line;
	}
}

void GdcApp8Set::Rectangle(int x, int y)
{
	COLOUR fill = c & 0xFF;

	fill = fill | (fill<<8) | (fill<<16) | (fill<<24);

#ifdef GDC_USE_ASM

// this duplicates the colour twice in eax to allow us to fill
// four pixels per write. this means we are using the whole
// 32-bit bandwidth to the video card :)

	uchar *p = Ptr;
	int Line = Dest->Line;

	if (x > 0 AND y > 0)
	{
		_asm {
			mov esi, p
			mov eax, fill
			mov edx, Line
		LoopY:	mov edi, esi
			add esi, edx
			mov ecx, x
			shr ecx, 2
			jz Odd
		LoopX:	mov [edi], eax
			add edi, 4
			dec ecx
			jnz LoopX
		Odd:	mov ecx, x
			and ecx, 3
			jz Next
		LoopO:	mov [edi], al
			inc edi
			dec ecx
			jnz LoopO
		Next:	dec y
			jnz LoopY
		}
	}
#else
	while (y--)
	{	
		int x4 = x/4;
		ulong *p = (ulong*) Ptr;

		for (int n=0; n<x4; n++) *p++ = fill;

		uchar *cp = (uchar*) p;
		for (int n=0; n<x%4; n++) *cp++ = c;

		Ptr += Dest->Line;
	}
#endif
}

class Node {
public:
	int Index;
	COLOUR Rgb;
	int Count;
	Node *a[8];

	Node()
	{
		Index = -1;
		Rgb = 0;
		Count = 0;
		for (int i=0; i<8; i++) a[i] = 0;
	}

	~Node()
	{
		for (int i=0; i<8; i++)
		{
			DeleteObj(a[i]);
		}
	}

	COLOUR Average()
	{
		if (Index < 0)
		{
			int R = 0;
			int G = 0;
			int B = 0;
			int Count = 0;

			for (int i=0; i<8; i++)
			{
				if (a[i])
				{
					COLOUR c = a[i]->Average();
					R += R24(c);
					G += G24(c);
					B += B24(c);
					Count++;
				}
			}

			if (Count)
			{
				Rgb = Rgb24(R/Count, G/Count, B/Count);
			}
			else
			{
				Rgb = 0;
			}
		}

		return Rgb;
	}
};

class OctTree {

	Node *Root;

public:
	OctTree()
	{
		Root = 0;
	}

	~OctTree()
	{
		DeleteObj(Root);
	}

	int BitIndex(COLOUR c, int Bit)
	{
		uchar r = (R24(c) >> Bit) & 1;
		uchar g = (G24(c) >> Bit) & 1;
		uchar b = (B24(c) >> Bit) & 1;
		return (r << 2) | (g << 1) | b;
	}

	void AddColour(COLOUR Index, COLOUR Rgb)
	{
		if (NOT Root)
		{
			Root = NEW(Node);
		}

		Node *Current = Root;
		for (int i=0; i<8; i++)
		{
			int n = BitIndex(Rgb, 7-i);
			if (NOT Current->a[n])
			{
				Current->a[n] = NEW(Node);
			}
			Current = Current->a[n];
		}

		Current->Index = Index;
		Current->Rgb = Rgb;
	}

	void ProcessBranches()
	{
		if (Root)
		{
			Root->Average();
		}
	}

	float Dist(COLOUR c1, COLOUR c2)
	{
		int r = (int) R24(c1) - R24(c2);
		int g = (int) G24(c1) - G24(c2);
		int b = (int) B24(c1) - B24(c2);
		return sqrt(r*r + g*g + b*b);
	}

	int NearestColour(COLOUR Rgb)
	{
		if (Root)
		{
			Node *Current = Root;
			int r1 = R24(Rgb);
			int g1 = G24(Rgb);
			int b1 = B24(Rgb);
			int r, g, b;
			COLOUR c2;

			for (int i=0; i<8 AND Current; i++)
			{
				int Bi = BitIndex(Rgb, 7-i);
				if (Current->a[Bi])
				{
					Current = Current->a[Bi];
				}
				else
				{
					int BestDist = 1000000;
					Node *Best = NULL;
					for (int n=0; n<8; n++)
					{
						if (Current->a[n])
						{
							c2 = Current->a[n]->Rgb;
							r = r1 - R24(c2);
							g = g1 - G24(c2);
							b = b1 - B24(c2);
							int CurDist = sqrt(r*r + g*g + b*b);

							if (CurDist < BestDist)
							{
								BestDist = CurDist;
								Best = Current->a[n];
							}
						}
					}
					Current = Best;
				}
			}

			return Current->Index;
		}

		return 0;
	}
};

uchar Div51[256] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5};

uchar Mod51[256] = {
	0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
	20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
	38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0,  1,  2,  3,  4,  5,  6,
	7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
	26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43,
	44, 45, 46, 47, 48, 49, 50, 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12,
	13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
	31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
	49, 50, 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17,
	18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35,
	36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0,  1,  2,  3,
	4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
	23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0};

uchar DitherIndex8[64] = {
	0,  38,  9, 47,  2, 40, 11, 50,
	25, 12, 35, 22, 27, 15, 37, 24,
	6,  44,  3, 41,  8, 47,  5, 43,
	31, 19, 28, 15, 34, 21, 31, 18,
	1,  39, 11, 49,  0, 39, 10, 48,
	27, 14, 36, 23, 26, 13, 35, 23,
	7,  46,  4, 43,  7, 45,  3, 42,
	33, 20, 30, 17, 32, 19, 29, 16};

uchar Mul8[8] = {0, 8, 16, 24, 32, 40, 48, 56};
uchar Mul6[6] = {0, 6, 12, 18, 24, 30};
uchar Mul36[6] = {0, 36, 72, 108, 144, 180};

bool GdcApp8Set::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	Dest->Flags &= ~GDC_UPDATED_PALETTE;
	if (Src)
	{
		switch (Src->Bits)
		{
			case 8:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemCpy(Ptr, s, Src->x);
					s += Src->Line;
					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				if (Pal)
				{
					for (int y=0; y<Src->y; y++)
					{
						ushort *s = (ushort*) (Src->Base + (y * Src->Line));
						uchar *d = Ptr;

						for (int x=0; x<Src->x; x++)
						{
							*d++ = Pal->MatchRgb(Rgb16To24(*s));
							s++;
						}

						Ptr += Dest->Line;
					}
				}
				break;
			}
			case 24:
			{
				switch (GdcD->GetOption(GDC_REDUCE_TYPE))
				{
					#ifdef DLN
					case REDUCE_DL1:
					{
						SURFACE in, out;
						uchar DlPal[3][256];

						in.x = Src->x;
						in.y = Src->y;
						in.line = Src->Line;
						in.bits = Src->Bits;
						in.base = Src->Base;

						memcpy(&out, &in, sizeof(in));
						out.base = Ptr;
						out.line = Dest->Line;
						out.bits = 8;

						dl1quant(&in, &out, 256, FALSE, DlPal);

						if (NOT Pal)
						{
							Pal = NEW(GdcPalette(0, 256));
						}
						
						if (Pal)
						{
							Dest->Flags |= GDC_UPDATED_PALETTE;
							for (int i=0; i<Pal->GetSize(); i++)
							{
								GdcRGB *p = (*Pal)[i];
								if (p)
								{
									p->R = DlPal[2][i];
									p->G = DlPal[1][i];
									p->B = DlPal[0][i];
								}
							}
						}
						break;
					}
					#endif
					case REDUCE_NEAREST:
					{
						if (NOT Pal) break;

						OctTree Tree;
						GdcRGB *rgb = (*Pal)[0];

						for (int i=0; i<Pal->GetSize(); i++, rgb++)
						{
							Tree.AddColour(i, Rgb24(rgb->R, rgb->G, rgb->B));
						}
						Tree.ProcessBranches();

						for (int y=0; y<Src->y; y++)
						{
							uchar *s = Src->Base + (y * Src->Line);
							uchar *d = Ptr;

							for (int x=0; x<Src->x; x++)
							{
								// *d++ = Tree.NearestColour(Rgb24(s[2], s[1], s[0]));
								*d++ = Pal->MatchRgb(Rgb24(s[2], s[1], s[0]));
								s += 3;
							}

							Ptr += Dest->Line;
						}
						break;
					}
					case REDUCE_HALFTONE:
					{
						for (int y=0; y<Src->y; y++)
						{
							uchar *S = Src->Base + (y * Src->Line);
							uchar *D = Ptr;
							int ym = y % 8;

							for (int x=0; x<Src->x; x++)
							{
								int JJJ = DitherIndex8[Mul8[x&7] + ym];
								int r = *S++;
								int g = *S++;
								int b = *S++;
								
								*D++ =	(Div51[r] + (Mod51[r] > JJJ)) +
			        					Mul6[Div51[g] + (Mod51[g] > JJJ)] +
			        					Mul36[Div51[b] + (Mod51[b] > JJJ)];
							}

							Ptr += Dest->Line;
						}
						break;
					}
					case REDUCE_ERROR_DIFFUSION:
					{
						// dist = sqrt(0.299*SQR(r2-r1) + 0.587*SQR(g2-g1) + 0.114*SQR(b2-b1))
						// sci.engr.color ???
						// http://www.inforamp.net/~poynton/Poynton-colour.html
						uchar GammaTable[256];
						double Gamma = 1.4;
						if (NOT Pal) break;

						for (int i=0; i<256; i++)
						{
							GammaTable[i] = (uchar) (pow(((double)i)/256, Gamma) * 256);
						}

						int Size = Src->x * 3;
						uchar *Memory = NEW(uchar[(Size*2)+12]);
						if (Memory)
						{
							OctTree Tree;
							GdcRGB *rgb = (*Pal)[0];

							for (int i=0; i<Pal->GetSize(); i++, rgb++)
							{
								Tree.AddColour(i, Rgb24(rgb->R, rgb->G, rgb->B));
							}
							Tree.ProcessBranches();

							uchar *Buffer[2];

							Buffer[0] = Memory + 3;
							Buffer[1] = Memory + Size + 9;

							memcpy(Buffer[0], Src->Base, Size);
							if (Src->y > 1)
							{
								memcpy(Buffer[1], Src->Base + Src->Line, Size);
							}

							for (int y=0; y<Src->y; y++)
							{
								uchar *S = Buffer[0];
								uchar *SNext = Buffer[1];
								uchar *D = Ptr;

								for (int x=0; x<Src->x; x++, D++, S+=3, SNext+=3)
								{
									// *D = Tree.NearestColour(Rgb24(S[2], S[1], S[0]));
									*D = Pal->MatchRgb(Rgb24(GammaTable[S[2]], GammaTable[S[1]], GammaTable[S[0]]));
									char *P = (char*) (*Pal)[*D];

									for (int i=0; i<3; i++)
									{
										int nError = S[2-i] - P[i];
										int nRemainder = 0;
										int nCarry;

										nCarry = nError * 7 / 16;
										nRemainder += nCarry;
										S[i+3] = limit(S[i+3]+nCarry, 0, 255);

										nCarry = nError * 3 / 16;
										nRemainder += nCarry;
										SNext[i-3] = limit(SNext[i-3]+nCarry, 0, 255);

										nCarry = nError * 5 / 16;
										nRemainder += nCarry;
										SNext[i] = limit(SNext[i]+nCarry, 0, 255);

										SNext[i+3] = limit(SNext[i+3]+nError-nRemainder, 0, 255);
									}
								}
								
								memcpy(Buffer[0], Buffer[1], Size);
								if (y < Src->y - 1)
								{
									memcpy(Buffer[1], Src->Base + ((y + 1) * Src->Line), Size);
								}

								Ptr += Dest->Line;
							}
						}
						break;
					}
				}
				break;
			}
		}
	}
	return TRUE;
}

// 8 bit or sub functions
void GdcApp8Or::Set()
{
	*Ptr |= c;
}

void GdcApp8Or::VLine(int height)
{
	while (height--)
	{
		*Ptr |= c;
		Ptr += Dest->Line;
	}
}

void GdcApp8Or::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) Ptr[n] |= c;
		Ptr += Dest->Line;
	}
}

bool GdcApp8Or::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		switch (Src->Bits)
		{
			case 8:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemOr(Ptr, s, Src->x);
					s += Src->Line;
					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	return TRUE;
}

// 8 bit AND sub functions
void GdcApp8And::Set()
{
	*Ptr &= c;
}

void GdcApp8And::VLine(int height)
{
	while (height--)
	{
		*Ptr &= c;
		Ptr += Dest->Line;
	}
}

void GdcApp8And::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) Ptr[n] &= c;
		Ptr += Dest->Line;
	}
}

bool GdcApp8And::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemAnd(Ptr, s, Src->x);
			s += Src->Line;
			Ptr += Dest->Line;
		}
	}
	return TRUE;
}

// 8 bit XOR sub functions
void GdcApp8Xor::Set()
{
	*Ptr ^= c;
}

void GdcApp8Xor::VLine(int height)
{
	while (height--)
	{
		*Ptr ^= c;
		Ptr += Dest->Line;
	}
}

void GdcApp8Xor::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) Ptr[n] ^= c;
		Ptr += Dest->Line;
	}
}

bool GdcApp8Xor::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemXor(Ptr, s, Src->x);
			s += Src->Line;
			Ptr += Dest->Line;
		}
	}
	return TRUE;
}
