/*
**	FILE:		File.h
**	AUTHOR:		Matthew Allen
**	DATE:		11/7/95
**	DESCRIPTION:	File subsystem header
**
**
**	Copyright (C) 1995, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#ifndef __FILE_H
#define __FILE_H

#include <StorageKit.h>

#include "device.h"
#include "memdev.h"

#define MAX_PATH				256

#define O_READ					B_READ_ONLY
#define O_WRITE					B_WRITE_ONLY
#define O_READWRITE				B_READ_WRITE

// Time/Date related functions infomation
#define DECNANOPERSEC				10000000
#define SECPERMINUTE				60
#define SECPERHOUR				(60 * SECPERMINUTE)
#define SECPERDAY				(24 * SECPERHOUR)

// File attributes
#define FA_NORMAL				0x0000
#define FA_READONLY				0x0001
#define FA_HIDDEN				0x0002
#define FA_SYSTEM				0x0004
#define FA_VOLUME				0x0008
#define FA_DIRECTORY				0x0010
#define FA_ARCHIVE				0x0020

class NodeInfo {
protected:
	char *Name;
	int Attributes;
	quad Size;

public:
	NodeInfo();
	NodeInfo(char *n, int a, quad s);
	virtual ~NodeInfo();

	virtual int operator ^(NodeInfo &N);
	char *GetName() { return Name; }
	int GetAttributes() { return Attributes; }
	quad GetSize() { return Size; }
};

class VDirectory {
public:
	VDirectory() { }
	virtual ~VDirectory() { Close(); }
	virtual void GetDirType(char *Str, int Len) {}
	virtual bool First(char *Name, char Must, char Search) { return TRUE; }
	virtual bool Next() { return TRUE; }
	virtual bool Close() { return TRUE; }

	virtual long GetAttributes() { return 0; }
	virtual char *GetName() { return NULL; }
	virtual void GetName(char *Buf, int Len) { }
	virtual const quad GetCreationTime() { return 0; }
	virtual const quad GetLastAccessTime() { return 0; }
	virtual const quad GetLastWriteTime() { return 0; }
	virtual const quad GetSize() { return 0; }
};

class VDirView : public VDirectory {

	int Items;
	int Files;
	int Dirs;
	char CurrentDir[MAX_PATH];

	class Node : public NodeInfo {

		Node *Left, *Right;

	public:
		Node();
		Node(char *n, int a, quad s);
		~Node();

		Node **Traverse(Node **ppNode, int Type);
		void Add(Node *Info);

	} *Root, **Index;

public:
	VDirView();
	~VDirView();

	bool Read(char *Dir);

	int GetFiles() { return Files; }
	NodeInfo *File(int i)
	{
		if (Index) return (NodeInfo*) Index[Dirs + i];
		return 0;
	}

	int GetDirs() { return Dirs; }
	NodeInfo *Dir(int i)
	{
		if (Index) return (NodeInfo*) Index[i];
		return 0;
	}

	int GetItems() { return Items; }
	NodeInfo *operator [](int i)
	{
		if (Index) return (NodeInfo*) Index[i];
		return 0;
	}
};

class VBeDir : public VDirView {

	BDirectory *BDir;
	BEntry Entry;

public:
	VBeDir();
	VBeDir(char *Path);
	~VBeDir();

	bool First(char *Name, char Must, char Search);
	bool Next();
	bool Close();

	long GetAttributes();
	char *GetName();
	void GetName(char *Buf, int Len);
	const quad GetCreationTime();
	const quad GetLastAccessTime();
	const quad GetLastWriteTime();
	const quad GetSize();
};

// classes
class VolumeInfo {
public:
	char		Drive;
	char		Name[32];
	char		System[32];
	ulong		MaxPath;
	ulong		Flags;
};

#define DIF_NONE				0
#define DIF_3_5FLOPPY				1
#define DIF_5_25FLOPPY				2
#define DIF_HARDDISK				3
#define DIF_CDROM				4
#define DIF_NET					5
#define DIF_RAM					6
#define DIF_REMOVABLE				7
#define DIF_MAX					9

class DriveInfo {
public:
	int		Type;
	char		Name[64];
	long		TotalSpace;
	long		FreeSpace;
};

#define FileDev			(FileDevice::GetInstance())

class FileDevice : public GDevice {

	friend class VFile;

	int NumDrive;
	DriveInfo *DriveList;
	int FillDiskList();

	static FileDevice *Instance;

public:
	FileDevice();
	~FileDevice();
	static FileDevice *GetInstance() { return Instance; }

	int GetNumDrives() { return NumDrive; }
	DriveInfo *GetDriveInfo(int Drive);
	VDirectory *GetDir(char *DirName);

	bool CreateDirectory(char *PathName);
	bool RemoveDirectory(char *PathName);
};

class VFile : public BFile {
protected:
	char	*Name;
	bool	Swap;
	bool	Status;

	int SwapRead(uchar *Buf, int Size);
	int SwapWrite(uchar *Buf, int Size);

public:
	VFile();
	virtual ~VFile();

	virtual bool Open(char *Name, int Attrib);
	virtual bool Open();
	virtual bool Close();

	virtual int Read(void *Buffer, int Size);
	virtual int Write(void *Buffer, int Size);
	virtual int Seek(int To, int Whence);
	virtual int GetPosition();
	virtual int GetSize();
	virtual bool SetSize(int Size);
	virtual bool Eof();
	void SetStatus(bool s);
	bool GetStatus();

	// swap bytes on variable io
	void SetSwap(bool s) { Swap = s; }
	bool GetSwap() { return Swap; }

	int ReadStr(char *Buf, int Size);
	int WriteStr(char *Buf, int Size);

	int Print(char *Format, ...);

	// Read
	virtual VFile &operator >> (char		&i);
	virtual VFile &operator >> (signed char		&i);
	virtual VFile &operator >> (unsigned char	&i);
	virtual VFile &operator >> (signed short	&i);
	virtual VFile &operator >> (unsigned short	&i);
	virtual VFile &operator >> (signed int		&i);
	virtual VFile &operator >> (unsigned int	&i);
	virtual VFile &operator >> (signed long		&i);
	virtual VFile &operator >> (unsigned long	&i);
	virtual VFile &operator >> (float		&i);
	virtual VFile &operator >> (double		&i);
	virtual VFile &operator >> (long double		&i);
	virtual VFile &operator >> (quad		&i);

	// Write
	virtual VFile &operator << (char		i);
	virtual VFile &operator << (signed char		i);
	virtual VFile &operator << (unsigned char	i);
	virtual VFile &operator << (signed short	i);
	virtual VFile &operator << (unsigned short	i);
	virtual VFile &operator << (signed int		i);
	virtual VFile &operator << (unsigned int	i);
	virtual VFile &operator << (signed long		i);
	virtual VFile &operator << (unsigned long	i);
	virtual VFile &operator << (float		i);
	virtual VFile &operator << (double		i);
	virtual VFile &operator << (long double		i);
	virtual VFile &operator << (quad		i);
};

class VFileBuf : public VFile {

	int BufPos;
	int BufSize;
	uchar *Data;

	int SwapRead(uchar *Buf, int Size);
	int SwapWrite(uchar *Buf, int Size);

public:
	VFileBuf();
	virtual ~VFileBuf();

	virtual bool Open(char *Name, int Attrib);
	virtual bool Open();
	virtual bool Close();

	virtual int Read(void *Buffer, int Size);
	virtual int Write(void *Buffer, int Size);
	virtual int Seek(int To, int Whence);
	virtual int GetPosition();
	virtual int GetSize();
	virtual bool SetSize(int Size);
	virtual bool Eof();
	int ReadStr(char *Buf, int Size);
	int WriteStr(char *Buf, int Size);

	VFile &operator >> (           char & i );
	VFile &operator >> (    signed char & i );
	VFile &operator >> (  unsigned char & i );
	VFile &operator >> (   signed short & i );
	VFile &operator >> ( unsigned short & i );
	VFile &operator >> (     signed int & i );
	VFile &operator >> (   unsigned int & i );
	VFile &operator >> (    signed long & i );
	VFile &operator >> (  unsigned long & i );
	VFile &operator >> (          float & i );
	VFile &operator >> (         double & i );
	VFile &operator >> (    long double & i );

	VFile &operator << (                char  i );
	VFile &operator << (         signed char  i );
	VFile &operator << (       unsigned char  i );
	VFile &operator << (        signed short  i );
	VFile &operator << (      unsigned short  i );
	VFile &operator << (          signed int  i );
	VFile &operator << (        unsigned int  i );
	VFile &operator << (         signed long  i );
	VFile &operator << (       unsigned long  i );
	VFile &operator << (               float  i );
	VFile &operator << (              double  i );
	VFile &operator << (         long double  i );
};

// C functions
extern bool Match(char *Name, char *Mask);
extern bool ResolveShortcut(char *LinkFile, char *Path, int Length);
extern bool FileExists(char *Name);
extern int SizeofStr(char *s);
extern void WriteStr(VFile &f, char *s);
extern char *ReadStr(VFile &f);

#endif
