/*hdr
**      FILE:           GuiProgress.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           11/11/98
**      DESCRIPTION:    Progress stuff
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <time.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

/////////////////////////////////////////////////////////////////////////////////////////
ProgressList::ProgressList()
{
	InUse = FALSE;
}

bool ProgressList::Lock()
{
	while (InUse)
	{
		snooze(20);
	}

	InUse = TRUE;

	return TRUE;
}

void ProgressList::Unlock()
{
	InUse = FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////
Progress::Progress()
{
	Description = 0;
	Start = 0;
	Val = Low = 0;
	High = 1;
	Type = 0;
	Scale = 1.0;
}

Progress::Progress(char *desc, int l, int h, char *type, double scale)
{
	Description = NewStr(desc);
	Start = 0;
	Val = Low = l;
	High = h;
	Type = type;
	Scale = scale;
}

Progress::~Progress()
{
	DeleteArray(Description);
}

void Progress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
}

void Progress::GetLimits(int *l, int *h)
{
	if (l) *l = Low;
	if (h) *h = High;
}

void Progress::SetDescription(char *d)
{
	DeleteArray(Description);
	Description = NewStr(d);
}

Progress &Progress::operator =(Progress &p)
{
	SetDescription(p.GetDescription());

	int h, l;
	p.GetLimits(&h, &l);
	SetLimits(h, l);

	SetScale(p.GetScale());
	SetType(p.GetType());
	Value(p.Value());

	return *this;
}

/////////////////////////////////////////////////////////////////////////////////////////
#define IDC_DESCRIPTION				100
#define IDC_VALUE				101
#define IDC_RATE				102
#define IDC_PROGRESS				103
#define IDC_BUTTON				104

#define PANE_X					220
#define PANE_Y					120

GProgressPane::GProgressPane()
{
	Pos.ZOff(210, 100);
	GWindow::Name("Progress");
	Canceled = FALSE;
	Wait = FALSE;
	Ref = 0;

	Children.Insert(Desc	= NEW(GText(IDC_DESCRIPTION, 10, 10, 150, 20, "")));
	Children.Insert(ValText	= NEW(GText(IDC_VALUE, 10, 35, 90, 20, "")));
	Children.Insert(Rate	= NEW(GText(IDC_RATE, 105, 35, 90, 20, "")));
	Children.Insert(Bar	= NEW(GProgress(IDC_PROGRESS, 10, 60, 200, 15, "Progress")));
	Children.Insert(But	= NEW(GButton(IDC_BUTTON, 70, 85, 80, 20, "Request Abort")));
}

GProgressPane::~GProgressPane()
{
}

void GProgressPane::SetLimits(int l, int h)
{
	Progress::SetLimits(l, h);
	if (Bar)
	{
		Bar->SetLimits(0, 30000);
	}
}

void GProgressPane::Value(int v)
{
	char Str[256];

	Progress::Value(v);
	if (Start == 0)
	{
		// initialize the clock
		Start = clock();
	}
	else if (Rate)
	{
		// calc rate
		int Now = clock();
		double Secs = ((double) Now - Start) / (double) CLOCKS_PER_SEC;
		double PerSec;
		
		if (Secs != 0.0)
		{
			PerSec = ((double) Val - Low) / Secs;
		}
		else
		{
			PerSec = 0;
		}
		
		sprintf(Str, "@ %.2f %s / sec", PerSec * Scale, (Type) ? Type : "");
		Rate->GControl::Name(Str);
	}

	if (ValText)
	{
		sprintf(Str, "%.0f of %.0f %s", Val * Scale, (High - Low) * Scale, (Type) ? Type : "");
		ValText->GControl::Name(Str);
	}

	if (Bar)
	{
		if (High != Low)
		{
			double Raw = ((double) v - Low) / ((double) High - Low);
			Bar->Value(Raw * 30000);
		}
		else
		{
			Bar->Value(0);
		}
	}
}

void GProgressPane::OnCreate()
{
	for (GWindow *w = Children.First(); w; w = Children.Next())
	{
		w->Attach(this);
	}
}

int GProgressPane::OnNotify(GControl *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDC_BUTTON:
		{
			Canceled = TRUE;
			if (Canceled AND GWindow::Parent AND NOT Wait)
			{
				delete GWindow::Parent;
			}

			if (But)
			{
				But->GControl::Name("Waiting...");
			}

			break;
		}
	}

	return 0;
}

void GProgressPane::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	DrawMoveShallowEdge(pDC, r, RAISED);
	pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
	pDC->Rectangle(&r);
}

bool GProgressPane::Pour(GdcComplexRegion &r)
{
	GdcRegion *Best = FindLargest(r);
	if (Best)
	{
		Pos.ZOff(PANE_X, PANE_Y);
		Pos.Offset(Best->x1-Pos.x1, Best->y1-Pos.y1);
		SetPos(Pos, TRUE);
		return TRUE;
	}
	return FALSE;
}

GFont GProgressPane::GetFont()
{
	// GdcBeTtf *Fnt = SysFont;
	// return (Fnt) ? Fnt->Handle() : 0;
	return GFont();
}

char *GProgressPane::GetDescription()
{
	return Progress::GetDescription();
}

void GProgressPane::SetDescription(char *d)
{
	Progress::SetDescription(d);
	if (Desc)
	{
		Desc->GControl::Name(d);
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
GProgressDlg::GProgressDlg(GWindow *parent, bool wait)
{
	Wait = wait;
	GWindow::Parent = parent;

	#ifdef WIN32
	Pos.ZOff(	(GetSystemMetrics(SM_CXDLGFRAME) * 2) + 2,
				(GetSystemMetrics(SM_CYDLGFRAME) * 2) +
				GetSystemMetrics(SM_CYCAPTION) - 1);
	#else if BEOS
	Pos.ZOff(30, 10);
	#endif

	Pos.x2 += PANE_X;
	Pos.y2 += PANE_Y * 2;
	Pos.Offset(400, 400);
	MoveToCenter();

	GWindow::Name("Progress");
	DoModeless();
}

GProgressDlg::~GProgressDlg()
{
}

void GProgressDlg::OnCreate()
{
	Push();
}

int GProgressDlg::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
	}

	return GDialog::OnEvent(Msg);
}

GProgressPane *GProgressDlg::Push()
{
	GProgressPane *Pane = NEW(GProgressPane);
	if (Pane)
	{
		Pane->Wait = Wait;
		Pane->Attach(this);
		Progri.Insert(Pane);

		int y = 0;
		for (GProgressPane *p = Progri.First(); p; p = Progri.Next())
		{
			GdcRegion r(0, y, PANE_X, y + PANE_Y);
			p->SetPos(r);
			y = r.y2 + 1;
		}

		GWindow::Invalidate();

		/*
		GdcRegion r;
		r.ZOff(	(GetSystemMetrics(SM_CXFIXEDFRAME) * 2),
			(GetSystemMetrics(SM_CYFIXEDFRAME) * 2) + GetSystemMetrics(SM_CYCAPTION));

		r.x2 += PANE_X;
		r.y2 += PANE_Y * Progri.GetItems();
		r.Offset(Pos.x1, Pos.y1);

		SetPos(r, TRUE);
		*/
	}

	return Pane;
}

void GProgressDlg::Pop(GProgressPane *p)
{
	GProgressPane *Pane = (p) ? p : Progri.Last();
	if (Pane)
	{
		Pane->Detach();
		Progri.Delete(Pane);
		GWindow::Invalidate();

		DeleteObj(Pane);

		int y = 0;
		for (GProgressPane *p = Progri.First(); p; p = Progri.Next())
		{
			GdcRegion r(0, y, PANE_X, y + PANE_Y);
			p->SetPos(r);
			y = r.y2 + 1;
		}

		GWindow::Invalidate();

		/*
		GdcRegion r;
		r.ZOff(	(GetSystemMetrics(SM_CXBORDER) * 2),
			(GetSystemMetrics(SM_CYBORDER) * 2) + GetSystemMetrics(SM_CYCAPTION));

		r.x2 += PANE_X;
		r.y2 += PANE_Y * Progri.GetItems();
		r.Offset(Pos.x1, Pos.y1);

		SetPos(r);
		*/
	}
}

char *GProgressDlg::GetDescription()
{
	GProgressPane *Pane = Progri.First();
	return (Pane) ? Pane->GetDescription() : 0;
}

void GProgressDlg::SetDescription(char *d)
{
	GProgressPane *Pane = Progri.First();
	if (Pane)
	{
		Pane->SetDescription(d);
	}
}

void GProgressDlg::GetLimits(int *l, int *h)
{
	GProgressPane *Pane = Progri.First();
	if (Pane) Pane->GetLimits(l, h);
	else
	{
		if (l) *l = 0;
		if (h) *h = 0;
	}
}

void GProgressDlg::SetLimits(int l, int h)
{
	GProgressPane *Pane = Progri.First();
	if (Pane) Pane->SetLimits(l, h);
}

int GProgressDlg::Value()
{
	GProgressPane *Pane = Progri.First();
	return (Pane) ? Pane->Value() : 0;
}

void GProgressDlg::Value(int v)
{
	GProgressPane *Pane = Progri.First();
	if (Pane) Pane->Value(v);
}

bool GProgressDlg::IsCanceled()
{
	GProgressPane *Pane = Progri.First();
	return (Pane) ? Pane->IsCanceled() : 0;
}

char *GProgressDlg::GetType()
{
	GProgressPane *Pane = Progri.First();
	return (Pane) ? Pane->GetType() : 0;
}

void GProgressDlg::SetType(char *t)
{
	GProgressPane *Pane = Progri.First();
	if (Pane) Pane->SetType(t);
}

void GProgressDlg::OnPaint(GdcBasePrimitives *pDC)
{
	pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
	pDC->Rectangle();
}

// Sync this window with the contents of a list.
void GProgressDlg::OnSync(ProgressList *Prg)
{
	if (Prg AND Prg->Lock())
	{
		GProgressPane *Pane = Progri[1];

		for (Progress *p = Prg->First(); p; p = Prg->Next())
		{
			bool NewPane = FALSE;
			if (Pane)
			{
				if (Pane->Ref != p)
				{
					Pop(Pane);
					NewPane = TRUE;
				}
				else
				{
					// pane already exist for this progress
					// so just update the value
					if (Pane->Value() != p->Value())
					{
						Pane->Value(p->Value());
					}
				}
			}
			else
			{
				NewPane = TRUE;
			}

			if (NewPane)
			{
				// new pane
				Pane = Push();
				if (Pane)
				{
					Pane->Ref = p;
					Pane->SetDescription(p->GetDescription());

					int h, l;
					p->GetLimits(&h, &l);
					Pane->SetLimits(h, l);

					Pane->Progress::SetScale(p->GetScale());
					Pane->SetType(p->GetType());
					Pane->Value(p->Value());

				}
			}

			Pane = Progri.Next();
		}

		// too many panes, delete some
		while (Progri.GetItems() > Prg->GetItems() + 1)
		{
			Pop();
		}

		Prg->Unlock();
	}
}
