/*hdr
**      FILE:           Store.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           29/10/98
**      DESCRIPTION:    Tree storage
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "defs.h"
#include "store.h"

///////////////////////////////////////////////////////////////////////////
#define STORAGE_ITEM_MAGIC		0x123400AA

struct StorageItemHeader {

	int Magic;
	int Type;
	int ObjSize;
	int AllocatedSize;
	
	int Next;
	int Prev;
	int Parent;
	int Child;

};

StorageItem::StorageItem()
{
	StoreType = 0;
	StoreLoc = 0;
	StoreNext = 0;
	StorePrev = 0;
	StoreParent = 0;
	StoreChild = 0;
	StoreSize = 0;
	StoreAllocatedSize = 0;

	Tree = 0;
	Next = 0;
	Prev = 0;
	Parent = 0;
	Child = 0;
	Object = 0;
}

StorageItem::~StorageItem()
{
	if (Object)
	{
		if (Object->StoreDirty)
		{
			VFile *f = 0;
			if (Tree)
			{
				f = &Tree->File;
				if (f)
				{
					f->Seek(StoreLoc, SEEK_SET);
					Serialize(*f, TRUE);
					Object->Serialize(*f, TRUE);
				}
			}
		}

		Object->Store = 0;
		DeleteObj(Object);
	}

	if (Parent)
	{
		Parent->Child = 0;
	}

	if (Prev)
	{
		Prev->Next = 0;
	}

	DeleteObj(Child);
	DeleteObj(Next);
}

bool StorageItem::WriteHeader(VFile &f)
{
	StorageItemHeader ItemHeader;

	f.Seek(StoreLoc, SEEK_SET);
	ItemHeader.Magic = STORAGE_ITEM_MAGIC;
	ItemHeader.Type = StoreType;
	ItemHeader.ObjSize = StoreSize;
	ItemHeader.AllocatedSize = StoreAllocatedSize;
	ItemHeader.Next = StoreNext;
	ItemHeader.Prev = StorePrev;
	ItemHeader.Parent = StoreParent;
	ItemHeader.Child = StoreChild;

	return (f.Write(&ItemHeader, sizeof(ItemHeader)) == sizeof(ItemHeader));
}

bool StorageItem::MoveToLoc(int NewLoc)
{
	bool Status = FALSE;
	if (NewLoc > 0)
	{
		if (Tree)
		{
			int SrcLoc = StoreLoc + sizeof(StorageItemHeader);
			int DestLoc = NewLoc + sizeof(StorageItemHeader);
			StorageItem *Item;

			VFile &File = Tree->File;
			int Type = StoreType;
			Status = TRUE;

			// turn current object into free space
			StoreType = -1; // free space
			Status &= WriteHeader(File);
			StoreType = Type;
			
			// get new location
			StoreLoc = NewLoc;

			// update surrounding pointers
			Item = GetParent();
			if (Item)
			{
				Item->StoreChild = StoreLoc;
				Status &= Item->Serialize(File, TRUE);
			}

			Item = GetChild();
			if (Item)
			{
				Item->StoreParent = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			Item = GetNext();
			if (Item)
			{
				Item->StorePrev = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			Item = GetPrev();
			if (Item)
			{
				Item->StoreNext = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			if (Status)
			{
				Status &= WriteHeader(File);
				if (Status)
				{
					int DataSize = StoreSize;
					int BufSize = 64 << 10;
					uchar *Buf = NEW(uchar[BufSize]);
					if (Buf)
					{
						while (DataSize > 0)
						{
							int r = min(DataSize, BufSize);
							File.Seek(SrcLoc, SEEK_SET);
							r = File.Read(Buf, r);
							File.Seek(DestLoc, SEEK_SET);
							File.Write(Buf, r);

							SrcLoc += r;
							DestLoc += r;
							DataSize -= r;
						}

						DeleteArray(Buf);
					}
				}
			}
		}
	}
	return Status;
}

VFile *StorageItem::GotoObject()
{
	VFile *f = 0;
	
	if (Tree)
	{
		f = &Tree->File;
		if (f)
		{
			f->Seek(StoreLoc + sizeof(StorageItemHeader), SEEK_SET);
		}
	}

	return f;
}

bool StorageItem::Serialize(VFile &f, bool Write)
{
	bool Status = FALSE;
	StorageItemHeader ItemHeader;

	if (Write)
	{
		if (SetSize(Sizeof()))
		{
			f.Seek(StoreLoc, SEEK_SET);
			if (f.GetPosition() == StoreLoc)
			{
				ItemHeader.Magic = STORAGE_ITEM_MAGIC;
				ItemHeader.Type = StoreType;
				ItemHeader.ObjSize = StoreSize;
				ItemHeader.AllocatedSize = StoreAllocatedSize;
				ItemHeader.Next = StoreNext;
				ItemHeader.Prev = StorePrev;
				ItemHeader.Parent = StoreParent;
				ItemHeader.Child = StoreChild;

				if (f.Write(&ItemHeader, sizeof(ItemHeader)) == sizeof(ItemHeader))
				{
					Status = TRUE;
				}
			}
		}
	}
	else
	{
		f.Seek(StoreLoc, SEEK_SET);
		if (f.GetPosition() == StoreLoc)
		{
			if (	f.Read(&ItemHeader, sizeof(ItemHeader)) == sizeof(ItemHeader))
			{
				if (ItemHeader.Magic == STORAGE_ITEM_MAGIC)
				{
					StoreType = ItemHeader.Type;
					StoreSize = ItemHeader.ObjSize;
					StoreAllocatedSize = ItemHeader.AllocatedSize;
					StoreNext = ItemHeader.Next;
					StorePrev = ItemHeader.Prev;
					StoreParent = ItemHeader.Parent;
					StoreChild = ItemHeader.Child;

					Status = TRUE;
				}
				#ifdef _DEBUG
				else
				{
					char Str[256];
					sprintf(Str, "Load Node failed at: ", StoreLoc);
					MessageBox(NULL, Str, "Debug", MB_OK);
				}
				#endif
			}
		}
	}

	return Status;
}

bool StorageItem::SetSize(int NewSize)
{
	bool Status = TRUE;

	if (NewSize > StoreAllocatedSize)
	{
		if (Tree)
		{
			StorageItemHeader ItemHeader;
			StorageItem *Item;

			VFile &File = Tree->File;
			int Type = StoreType;

			// turn current object into free space
			StoreType = -1; // free space
			Status &= WriteHeader(File);
			StoreType = Type;
			
			// get new location
			StoreLoc = File.GetSize();

			// update surrounding pointers
			Item = GetParent();
			if (Item)
			{
				Item->StoreChild = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			Item = GetChild();
			if (Item)
			{
				Item->StoreParent = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			Item = GetNext();
			if (Item)
			{
				Item->StorePrev = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			Item = GetPrev();
			if (Item)
			{
				Item->StoreNext = StoreLoc;
				Status &= Item->WriteHeader(File);
			}

			if (Status)
			{
				StoreSize = NewSize;
				StoreAllocatedSize = NewSize;
				Status &= Serialize(File, TRUE);
				if (Status AND Object)
				{
					File.Seek(StoreLoc + sizeof(StorageItemHeader), SEEK_SET);
					Status &= Object->Serialize(File, TRUE);
				}
			}
		}
		else
		{
			Status = FALSE;
		}
	}

	StoreSize = NewSize;
	return Status;
}

StorageItem *StorageItem::GetNext()
{
	if (NOT Next)
	{
		Next = Tree->LoadLocation(StoreNext);
		if (Next)
		{
			Next->Prev = this;
		}
		#ifdef _DEBUG
		else if (StoreNext)
		{
			// a bug!
			// patch the next pointer to null.
			StoreNext = 0;
			if (Tree)
			{
				WriteHeader(Tree->File);
			}

			char Str[256];
			sprintf(Str, "Next pointer invalid in Obj loc: %i", StoreLoc);
			MessageBox(NULL, Str, "Debug", MB_OK);
		}
		#endif
	}

	return Next;
}

StorageItem *StorageItem::GetPrev()
{
	if (NOT Prev)
	{
		Prev = Tree->LoadLocation(StorePrev);
		if (Prev)
		{
			Prev->Next = this;
		}
	}

	return Prev;
}

StorageItem *StorageItem::GetParent()
{
	if (NOT Parent)
	{
		Parent = Tree->LoadLocation(StoreParent);
		if (Parent)
		{
			Parent->Child = this;
		}
	}

	return Parent;
}

StorageItem *StorageItem::GetChild()
{
	if (NOT Child)
	{
		Child = Tree->LoadLocation(StoreChild);
		if (Child)
		{
			Child->Parent = this;
		}
		#ifdef _DEBUG
		else if (StoreChild)
		{
			// a bug!
			// patch the next pointer to null.
			StoreChild = 0;
			if (Tree)
			{
				WriteHeader(Tree->File);
			}

			char Str[256];
			sprintf(Str, "Child pointer invalid in Obj loc: %i", StoreLoc);
			MessageBox(NULL, Str, "Debug", MB_OK);
		}
		#endif
	}

	return Child;
}

StorageItem *StorageItem::CreateNext(StorageObj *Obj)
{
	if (Tree AND Obj)
	{
		Next = GetNext();
		if (Next)
		{
			Tree->DeleteItem(Next);
			DeleteObj(Next);
		}

		Next = Tree->CreateItem(Obj);
		if (Next)
		{
			// link this object to the next object
			StoreNext = Next->StoreLoc;
			Next->StorePrev = StoreLoc;
			Next->Prev = this;

			// save myself
			Serialize(Tree->File, TRUE);
			
			// save next
			Next->Serialize(Tree->File, TRUE);
		}
	}
	return Next;
}

StorageItem *StorageItem::CreateChild(StorageObj *Obj)
{
	if (Tree AND Obj)
	{
		Child = GetChild();
		if (Child)
		{
			Tree->DeleteItem(Child);
			DeleteObj(Child);
		}

		Child = Tree->CreateItem(Obj);
		if (Child)
		{
			// link this object to the Child object
			StoreChild = Child->StoreLoc;
			Child->StoreParent = StoreLoc;
			Child->Parent = this;

			// save myself
			Serialize(Tree->File, TRUE);
			
			// save Child
			Child->Serialize(Tree->File, TRUE);
		}
	}
	return Child;
}

StorageItem *StorageItem::CreateSub(StorageObj *Obj)
{
	StorageItem *NewItem = GetChild();
	if (NewItem)
	{
		while (NewItem AND NewItem->StoreNext)
		{
			NewItem = NewItem->GetNext();
		}

		if (NewItem)
		{
			NewItem = NewItem->CreateNext(Obj);
		}
	}
	else
	{
		NewItem = CreateChild(Obj);
	}
	return NewItem;
}

int StorageItem::Sizeof()
{
	return (Object) ? Object->Sizeof() : 0;
}

///////////////////////////////////////////////////////////////////////////
#define STORAGE_MAGIC			0x123400FF

struct StorageHeader {

	int Magic;
	int Used;
	int Free;
	int FirstLoc;
	int Objects;

	int Reserved[11];
};

StorageKit::StorageKit(char *filename)
{
	Status = FALSE;
	StoreLoc = 0;
	SpaceUsed = 0;
	SpaceFree = 0;
	Root = 0;
	Objects = 0;

	FileName = filename;
	if (File.Open(FileName, O_READWRITE))
	{
		Status = Serialize(File, File.GetSize() == 0);
	}
}

StorageKit::~StorageKit()
{
	DeleteObj(Root);

	File.Close();
	Status = FALSE;
}

StorageItem *StorageKit::GetRoot()
{
	if (NOT Root)
	{
		Root = LoadLocation(StoreLoc);
	}

	return Root;
}

StorageItem *StorageKit::CreateRoot(StorageObj *Obj)
{
	StorageItem *OldRoot = GetRoot();
	if (OldRoot)
	{
		DeleteItem(OldRoot);
	}

	Compact();
	Root = CreateItem(Obj);
	if (Root)
	{
		Objects = 1;
		StoreLoc = Root->StoreLoc;
		Serialize(File, TRUE);
	}
	
	return Root;
}

StorageItem *StorageKit::CreateItem(StorageObj *Obj)
{
	StorageItem *Item = NEW(StorageItem);
	if (Item)
	{
		// set items new location: at the end of the file
		Item->Tree = this;
		Item->StoreType = Obj->Type();
		Item->StoreLoc = File.GetSize();
		Item->Object = Obj;
		Item->StoreSize = Item->StoreAllocatedSize = Item->Sizeof();
		Obj->StoreDirty = TRUE;

		// write the item to the disk
		Item->Serialize(File, TRUE);

		// write the object to disk for the first time
		VFile *f = Item->GotoObject();
		if (f)
		{
			Item->Object->Serialize(*f, TRUE);
		}
	}
	return Item;
}

bool StorageKit::SeparateItem(StorageItem *Item)
{
	bool Status = FALSE;

	if (Item)
	{
		StorageItem *Parent = Item->GetParent();
		StorageItem *Next = Item->GetNext();
		StorageItem *Prev = Item->GetPrev();

		if (Parent)
		{
			if (Next)
			{
				// attach the parent to the next
				Parent->StoreChild = Next->StoreLoc;
				Parent->Child = Next;
				Next->StorePrev = 0;
				Next->Prev = 0;
				Next->StoreParent = Parent->StoreLoc;
				Next->Parent = Parent;
			}
			else
			{
				// we are an only child
				Parent->StoreChild = 0;
				Parent->Child = 0;
			}
		}
		else
		{
			if (Next)
			{
				if (Prev)
				{
					Next->StorePrev = Prev->StoreLoc;
					Next->Prev = Prev;
					Prev->StoreNext = Next->StoreLoc;
					Prev->Next = Next;
				}
				else
				{
					Next->StorePrev = 0;
					Next->Prev = 0;
				}
			}
			else if (Prev)
			{
				Prev->StoreNext = 0;
				Prev->Next = 0;
			}
		}

		if (Parent) Parent->Serialize(File, TRUE);
		if (Next) Next->Serialize(File, TRUE);
		if (Prev) Prev->Serialize(File, TRUE);

		Item->Next = 0;
		Item->StoreNext = 0;
		Item->Parent = 0;
		Item->StoreParent = 0;
		Item->Prev = 0;
		Item->StorePrev = 0;
		Status = Item->Serialize(File, TRUE);
	}

	return Status;
}

bool StorageKit::DeleteItem(StorageItem *Item)
{
	bool Status = TRUE;

	if (Item)
	{
		DeleteItem(Item->GetChild());
		DeleteItem(Item->GetNext());

		StorageItem *Prev = Item->GetPrev();
		if (Prev)
		{
			Prev->StoreNext = 0;
			Prev->Next = 0;
			Prev->Serialize(File, TRUE);
		}

		StorageItem *Parent = Item->GetParent();
		if (Parent)
		{
			Parent->StoreChild = 0;
			Parent->Child = 0;
			Parent->Serialize(File, TRUE);
		}

		Item->StoreType = 0;
		Status = Item->Serialize(File, TRUE);
	}

	return Status;
}

void StorageKit::AddItem(StorageItem *Item, List<Block> &Blocks)
{
	if (Item)
	{
		Block *New = NEW(Block);
		if (New)
		{
			New->Item = Item;
			New->Size = sizeof(StorageItemHeader) + Item->StoreAllocatedSize;

			if (Item->StoreAllocatedSize == 0 OR Item->StoreSize == 0)
			{
				if (Item->Object)
				{
					Item->StoreSize = Item->Object->Sizeof();
					Item->WriteHeader(File);
				}
			}

			Blocks.Insert(New);
		}

		AddItem(Item->GetChild(), Blocks);
		AddItem(Item->GetNext(), Blocks);
	}
}

bool StorageKit::Compact()
{
	bool Error = FALSE;
	List<Block> Blocks;
	List<Block> Sorted;
	StorageItem *Item = GetRoot();
	if (Item)
	{
		// get a list of all valid blocks
		AddItem(Item, Blocks);

		// sort the blocks
		while (Blocks.GetItems() > 0)
		{
			Block *Min = 0;
			for (Block *b = Blocks.First(); b; b = Blocks.Next())
			{
				if (NOT Min) Min = b;
				if (Min)
				{
					if (b->Item->StoreLoc < Min->Item->StoreLoc)
					{
						Min = b;
					}
				}
			}

			if (Min)
			{
				Blocks.Delete(Min);
				Sorted.Insert(Min);
			}
			else break;
		}

		// remove space between blocks
		for (Block *b = Sorted.First(); b;)
		{
			Block *Next = Sorted.Next();
			if (Next)
			{
				int EndOfCurrent = b->Item->StoreLoc + b->Size;
				if (Next->Item->StoreLoc > EndOfCurrent)
				{
					// gap found
					// need to move the Next->Item to EndOfCurrent
					if (NOT Next->Item->MoveToLoc(EndOfCurrent))
					{
						// somthing went wrong
						// lets just leave before we screw everything up
						Error = TRUE;
						break;
					}
				}
			}
			else
			{
				// last block so we truncate the file at the end of the block
				int EndOfCurrent = b->Item->StoreLoc + b->Size;
				File.SetSize(EndOfCurrent);
			}

			b = Next;
		}
	}

	for (Block *b = Sorted.First(); b; b = Sorted.Next())
	{
		DeleteObj(b);
	}
	
	return NOT Error;
}

StorageItem *StorageKit::LoadLocation(int Loc)
{
	StorageItem *Item = 0;

	if (Loc > 0 AND Status)
	{
		Item = NEW(StorageItem);
		if (Item)
		{
			Item->StoreLoc = Loc;
			if (Item->Serialize(File, FALSE))
			{
				Item->Tree = this;
			}
			else
			{
				DeleteObj(Item);
			}
		}
	}

	return Item;
}

bool StorageKit::Serialize(VFile &f, bool Write)
{
	bool Status = FALSE;
	StorageHeader Header;

	if (Write)
	{
		ZeroObj(Header);
		Header.Magic = STORAGE_MAGIC;
		Header.Used = SpaceUsed;
		Header.Free = SpaceFree;
		Header.FirstLoc = StoreLoc;
		Header.Objects = Objects;

		File.Seek(0, SEEK_SET);
		Status = File.Write(&Header, sizeof(Header)) == sizeof(Header);
	}
	else
	{
		File.Seek(0, SEEK_SET);
		Status = File.Read(&Header, sizeof(Header)) == sizeof(Header);
		if (Status)
		{
			Status = (Header.Magic == STORAGE_MAGIC);
			if (Status)
			{
				StoreLoc = Header.FirstLoc;
				SpaceUsed = Header.Used;
				SpaceFree = Header.Free;
				Objects = Header.Objects;
			}
		}
	}
	return Status;
}

bool StorageKit::ReparentItem(StorageItem *Item, StorageItem *NewParent)
{
	bool Status = FALSE;

	if (Item AND
		NewParent AND
		SeparateItem(Item))
	{
		StorageItem *Last = NewParent->GetChild();
		if (Last)
		{
			while (Last AND Last->StoreNext)
			{
				Last = Last->GetNext();
			}

			if (Last)
			{
				Last->StoreNext = Item->StoreLoc;
				Last->Next = Item;
				Item->StorePrev = Last->StoreLoc;
				Item->Prev = Last;

				Last->Serialize(File, TRUE);
			}
		}
		else
		{
			Item->StoreParent = NewParent->StoreLoc;
			Item->Parent = NewParent;
			NewParent->StoreChild = Item->StoreLoc;
			NewParent->Child = Item;

			NewParent->Serialize(File, TRUE);
		}

		Status = Item->Serialize(File, TRUE);
	}

	return Status;
}