/*hdr
**      FILE:           INet.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Internet sockets
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "Inet.h"

#ifdef BEOS
#define isdigit(c) ((c)>='0'AND(c)<='9')
#endif

////////////////////////////////////////////////////////////////////////////////////////////////
ISockets::ISockets()
{
	SocketsOpen = FALSE;
}

ISockets::~ISockets()
{
	#ifdef WIN32
	Close();
	#endif
}

bool ISockets::Open()
{
	#ifdef WIN32
	Close();

	uchar Major = 1;
	uchar Minor = 1;
	ushort Version = MAKEWORD(Major, Minor);
	WSADATA WsaData;

	// Start sockets
	int Result = WSAStartup(Version, &WsaData);
	if (NOT Result)
	{
		if (LOBYTE(WsaData.wVersion) == Major AND HIBYTE(WsaData.wVersion) == Minor)
		{
			SocketsOpen = Result == 0;
		}
		else
		{
			WSACleanup();
		}
	}

	return SocketsOpen;
	#else if BEOS
	return true;
	#endif
}

void ISockets::Close()
{
	#ifdef WIN32
	if (SocketsOpen)
	{
		WSACleanup();
		SocketsOpen = FALSE;
	}
	#endif
}


IConnection *ISockets::CreateConnection()
{
	IConnection *C = NULL;

	if (IsOpen())
	{
		C = NEW(IConnection);
		if (C)
		{
			C->Parent = this;
		}
	}

	return C;
}

////////////////////////////////////////////////////////////////////////////////////////////////
IConnection::IConnection()
{
	Connected = FALSE;
	LastError = 0;
	BytesWritten = 0;
	BytesRead = 0;
}

IConnection::~IConnection()
{
	Close();
}

bool IConnection::Open(char *HostAddr, int port)
{
	Close();

	Socket = socket(AF_INET, SOCK_STREAM, 0);
	if (Socket >= 0)
	{
		BytesWritten = 0;
		BytesRead = 0;

		if (isdigit(*HostAddr) AND strchr(HostAddr, '.'))
		{
			// Ip address
			IpAddress = inet_addr(HostAddr);
			Host = gethostbyaddr((const char *)&IpAddress, 4, AF_INET);
		}
		else
		{
			// Name address
			Host = gethostbyname(HostAddr);
		}

		if (NOT Host)
		{
			Error();
		}
		else
		{
			Port = port;
			RemoteAddr.sin_family = AF_INET;
			RemoteAddr.sin_port = htons(Port);
			memcpy(&RemoteAddr.sin_addr, Host->h_addr_list[0], sizeof(in_addr));

			if (connect(Socket, (sockaddr*) &RemoteAddr, sizeof(sockaddr)))
			{
				Error();
			}
			else
			{
				char Info[256];
				sprintf(Info,
						"Connect socket: %s [%i.%i.%i.%i], port: %i",
						HostAddr,
						(RemoteAddr.sin_addr.s_addr >> 24) & 0xFF,
						(RemoteAddr.sin_addr.s_addr >> 16) & 0xFF,
						(RemoteAddr.sin_addr.s_addr >> 8) & 0xFF,
						(RemoteAddr.sin_addr.s_addr) & 0xFF,
						Port);
				OnInfomation(Info);

				Connected = TRUE;
			}
		}

		if (NOT Connected)
		{
			closesocket(Socket);
		}
	}
	else
	{
		Error();
	}

	return Connected;
}

void IConnection::Close()
{
	if (Connected)
	{
		closesocket(Socket);
		Connected = FALSE;
		OnInfomation("Disconnect socket");
	}

	IpAddress = 0;
	Socket = -1;
}

bool IConnection::Write(char *Data, int Len, int Flags)
{
	bool Status = FALSE;
	if (Connected)
	{
		int Count = 0;
		while (Count < Len)
		{
			int Code = send(Socket, Data + Count, Len - Count, Flags);
			if (Code < 0)
			{
				Error();
				break;
			}
			else
			{
				Count += Code;
			}
		}

		Status = (Count == Len);
		if (Status)
		{
			BytesWritten += Len;
			OnWrite(Data, Len);
		}
	}

	return Status;
}

int IConnection::Read(char *Data, int Len, int Flags)
{
	if (Connected AND Data)
	{
		int Code = recv(Socket, Data, Len, Flags);
		if (Code < 0)
		{
			Error();
		}
		else
		{
			if (Code < Len)
			{
				// Just in case it's a string lets be safe
				*(Data + Code) = 0;
			}

			BytesRead += Code;
			OnRead(Data, Code);
		}

		return Code;
	}

	return 0;
}

int IConnection::ReadLen(char *Data, int Len, int Flags)
{
	int Pos = 0;

	if (Connected AND Data)
	{
		while (Pos < Len)
		{
			int Code = recv(Socket, Data + Pos, Len - Pos, Flags);

			if (Code < 0)
			{
				Error();
				break;
			}
			else
			{
				OnRead(Data + Pos, Code);
				Pos += Code;
			}
		}
	}

	return Pos;
}

int IConnection::Error()
{
	class ErrorMsg {
	public:
		int Code;
		char *Msg;
	} ErrorCodes[] = {
		#ifdef WIN32
		{WSAEACCES,		"Permission denied."},
		{WSAEADDRINUSE,		"Address already in use."},
		{WSAEADDRNOTAVAIL,	"Cannot assign requested address."},
		{WSAEAFNOSUPPORT,	"Address family not supported by protocol family."},
		{WSAEALREADY,		"Operation already in progress."},
		{WSAECONNABORTED,	"Software caused connection abort."},
		{WSAECONNREFUSED,	"Connection refused."},
		{WSAECONNRESET,		"Connection reset by peer."},
		{WSAEDESTADDRREQ,	"Destination address required."},
		{WSAEFAULT,			"Bad address."},
		{WSAEHOSTDOWN,		"Host is down."},
		{WSAEHOSTUNREACH,	"No route to host."},
		{WSAEINPROGRESS,	"Operation now in progress."},
		{WSAEINTR,			"Interrupted function call."},
		{WSAEINVAL,			"Invalid argument."},
		{WSAEISCONN,		"Socket is already connected."},
		{WSAEMFILE,			"Too many open files."},
		{WSAEMSGSIZE,		"Message too long."},
		{WSAENETDOWN,		"Network is down."},
		{WSAENETRESET,		"Network dropped connection on reset."},
		{WSAENETUNREACH,	"Network is unreachable."},
		{WSAENOBUFS,		"No buffer space available."},
		{WSAENOPROTOOPT,	"Bad protocol option."},
		{WSAENOTCONN,		"Socket is not connected."},
		{WSAENOTSOCK,		"Socket operation on non-socket."},
		{WSAEOPNOTSUPP,		"Operation not supported."},
		{WSAEPFNOSUPPORT,	"Protocol family not supported."},
		{WSAEPROCLIM,		"Too many processes."},
		{WSAEPROTONOSUPPORT,"Protocol not supported."},
		{WSAEPROTOTYPE,		"Protocol wrong type for socket."},
		{WSAESHUTDOWN,		"Cannot send after socket shutdown."},
		{WSAESOCKTNOSUPPORT,"Socket type not supported."},
		{WSAETIMEDOUT,		"Connection timed out."},
		{WSAEWOULDBLOCK,	"Resource temporarily unavailable."},
		{WSAHOST_NOT_FOUND,	"Host not found."},
		{WSANOTINITIALISED,	"Successful WSAStartup not yet performed."},
		{WSANO_DATA,		"Valid name, no data record of requested type."},
		{WSANO_RECOVERY,	"This is a non-recoverable error."},
		{WSASYSNOTREADY,	"Network subsystem is unavailable."},
		{WSATRY_AGAIN,		"Non-authoritative host not found."},
		{WSAVERNOTSUPPORTED,"WINSOCK.DLL version out of range."},
		{WSAEDISCON,		"Graceful shutdown in progress."},
		#endif
		{0, 0}};

	#ifdef WIN32
	LastError = WSAGetLastError();
	#endif
	ErrorMsg *Error = ErrorCodes;
	while (Error->Code AND Error->Code != LastError)
	{
		Error++;
	}

	OnError(LastError, (Error->Code) ? Error->Msg : "<unknown error>");
	return LastError;
}



