/*hdr
**      FILE:           Gif.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Gif file filter
**
**      Copyright (C) 1997-8, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"

// gif error codes
#define OUT_OF_MEMORY			-10
#define BAD_CODE_SIZE			-20
#define READ_ERROR			-1
#define WRITE_ERROR			-2
#define OPEN_ERROR			-3
#define CREATE_ERROR			-4

long code_mask[13] = {
	  0,
	  0x0001,
	  0x0003,
	  0x0007,
	  0x000F,
	  0x001F,
	  0x003F,
	  0x007F,
	  0x00FF,
	  0x01FF,
	  0x03FF,
	  0x07FF,
	  0x0FFF};

int GdcGif::get_byte()
{
	uchar c;
	
	*this >> c;
	if (NOT Eof()) return c;

	return READ_ERROR;
}

int GdcGif::out_line(uchar *pixels, int linewidth, int interlaced, int BitDepth)
{
	static int x, y;
	
	if (lines == 0)
	{
		x = 0;
		y = 0;
	}
			
	switch (BitDepth)
	{
		default:
		{
			for (int i=0; i<linewidth; i++, pixels++)
			{
				pDC->Colour(*pixels);
				pDC->Set(x++, y);

				if (x >= pDC->X())
				{
					x = 0;
					y++;
				}
			}
			break;
		}
		case 8:
		{
			for (int x=0; x<pDC->X(); x++)
			{	
				pDC->Colour(*pixels++);
				pDC->Set(x, lines);
			}
			break;
		}
		case 4:
		{
			for (int i=0; i<linewidth; i++, pixels++)
			{
				pDC->Colour(*pixels & 0xF);
				pDC->Set(x++, y);
			}
			
			if (lines % 2 == 1)
			{
				x = 0;
				y++;
			}
			break;
		}
		case 1:
		{
			for (int i=0; i<linewidth; i++, pixels++)
			{
				pDC->Colour((*pixels & 1) ? 0xFF : 0);
				pDC->Set(x++, y);
			}
			
			if (lines % 8 == 7)
			{
				x = 0;
				y++;
			}
			break;
		}
	}

	if (interlaced)
	{
		switch (pass)
		{
			case 0:
				lines += 8;
				if (lines >= pDC->Y())
				{
					lines = 4;
					pass++;
				}
				break;
				
			case 1:
				lines += 8;
				if (lines >= pDC->Y())
				{
					lines = 2;
					pass++;
				}
				break;

			case 2:
				lines += 4;
				if (lines >= pDC->Y())
				{
					lines = 1;
					pass++;
				}
				break;
				
			case 3:
				lines += 2;
				break;
		}
	}
	else
	{
		lines++;
	}

	return 0;
}

short GdcGif::init_exp(short size)
{
	curr_size = size + 1;
	top_slot = 1 << curr_size;
	clearc = 1 << size;
	ending = clearc + 1;
	slot = newcodes = ending + 1;
	navail_bytes = nbits_left = 0;

	return 0;
}

short GdcGif::get_next_code()
{
	short i, x;
	ulong ret;

	if (nbits_left == 0)
	{
		if (navail_bytes <= 0)
		{

			/* Out of bytes in current block, so read next block
			 */
			pbytes = byte_buff;
			if ((navail_bytes = get_byte()) < 0)
			{
				return(navail_bytes);
			}
			else if (navail_bytes)
			{
				for (i = 0; i < navail_bytes; ++i)
				{
					if ((x = get_byte()) < 0)
					{
						return(x);
					}

					byte_buff[i] = x;
				}
			}
		}

		b1 = *pbytes++;
		nbits_left = 8;
		--navail_bytes;
	}

	ret = b1 >> (8 - nbits_left);
	while (curr_size > nbits_left)
	{
		if (navail_bytes <= 0)
		{
			/* Out of bytes in current block, so read next block
			 */
			pbytes = byte_buff;

			if ((navail_bytes = get_byte()) < 0)
			{
				return(navail_bytes);
			}
			else if (navail_bytes)
			{
				for (i = 0; i < navail_bytes; ++i)
				{
					if ((x = get_byte()) < 0)
					{
						return(x);
					}

					byte_buff[i] = x;
				}
			}
		}

		b1 = *pbytes++;
		ret |= b1 << nbits_left;
		nbits_left += 8;
		--navail_bytes;
	}

	nbits_left -= curr_size;
	ret &= code_mask[curr_size];

	return ((short) ret);
}

short GdcGif::decoder(int BitDepth, uchar interlaced)
{
	uchar *sp, *bufptr;
	uchar *buf;
	short code, fc, oc, bufcnt;
	short c, size, ret;

	lines = 0;
	pass = 0;

	/* Initialize for decoding a new image...
	 */
	if ((size = get_byte()) < 0)
	{
		return(size);
	}

	if (size < 2 || 9 < size)
	{
		return(BAD_CODE_SIZE);
	}

	init_exp(size);

	/* Initialize in case they forgot to put in a clearc code.
	 * (This shouldn't happen, but we'll try and decode it anyway...)
	 */
	oc = fc = 0;

	/* Allocate space for the decode buffer
	 */
	buf = NEW(uchar[linewidth+1]);
	if (buf == NULL)
	{
		return (OUT_OF_MEMORY);
	}

	/* Set up the stack pointer and decode buffer pointer
	 */
	sp = stack;
	bufptr = buf;
	bufcnt = linewidth;

	/* This is the main loop.  For each code we get we pass through the
	 * linked list of prefix codes, pushing the corresponding "character" for
	 * each code onto the stack.  When the list reaches a single "character"
	 * we push that on the stack too, and then start unstacking each
	 * character for output in the correct order.  Special handling is
	 * included for the clearc code, and the whole thing ends when we get
	 * an ending code.
	 */
	while ((c = get_next_code()) != ending)
	{

		/* If we had a file error, return without completing the decode
		 */
		if (c < 0)
		{
			DeleteObj(buf);
			return(0);
		}

		/* If the code is a clearc code, reinitialize all necessary items.
		 */
		if (c == clearc)
		{
			curr_size = size + 1;
			slot = newcodes;
			top_slot = 1 << curr_size;

			/* Continue reading codes until we get a non-clearc code
			 * (Another unlikely, but possible case...)
			 */
			while ((c = get_next_code()) == clearc)
				;

			/* If we get an ending code immediately after a clearc code
			 * (Yet another unlikely case), then break out of the loop.
			 */
			if (c == ending)
			{
				break;
			}

			/* Finally, if the code is beyond the range of already set codes,
			 * (This one had better NOT happen...  I have no idea what will
			 * result from this, but I doubt it will look good...) then set it
			 * to color zero.
			 */
			if (c >= slot)
			{
				c = 0;
			}

			oc = fc = c;

			/* And let us not forget to put the char into the buffer... And
			 * if, on the off chance, we were exactly one pixel from the end
			 * of the line, we have to send the buffer to the out_line()
			 * routine...
			 */
			*bufptr++ = c;

			if (--bufcnt == 0)
			{
				if ((ret = out_line(buf, linewidth, interlaced, BitDepth)) < 0)
				{
					DeleteObj(buf);
					return (ret);
				}

				bufptr = buf;
				bufcnt = linewidth;
			}
		}
		else
		{

			/* In this case, it's not a clearc code or an ending code, so
			 * it must be a code code...  So we can now decode the code into
			 * a stack of character codes. (Clear as mud, right?)
			 */
			code = c;

			/* Here we go again with one of those off chances...  If, on the
			 * off chance, the code we got is beyond the range of those already
			 * set up (Another thing which had better NOT happen...) we trick
			 * the decoder into thinking it actually got the last code read.
			 * (Hmmn... I'm not sure why this works...  But it does...)
			 */
			if (code >= slot)
			{
				if (code > slot)
				{
					++bad_code_count;
				}

				code = oc;
				*sp++ = fc;
			}

			/* Here we scan back along the linked list of prefixes, pushing
			 * helpless characters (ie. suffixes) onto the stack as we do so.
			 */
			while (code >= newcodes)
			{
				*sp++ = suffix[code];
				code = prefix[code];
			}

			/* Push the last character on the stack, and set up the new
			 * prefix and suffix, and if the required slot number is greater
			 * than that allowed by the current bit size, increase the bit
			 * size.  (NOTE - If we are all full, we *don't* save the new
			 * suffix and prefix...  I'm not certain if this is correct...
			 * it might be more proper to overwrite the last code...
			 */
			*sp++ = code;
			if (slot < top_slot)
			{
				suffix[slot] = fc = code;
				prefix[slot++] = oc;
				oc = c;
			}

			if (slot >= top_slot)
			{
				if (curr_size < 12)
				{
					top_slot <<= 1;
					++curr_size;
				}
			}

			/* Now that we've pushed the decoded string (in reverse order)
			 * onto the stack, lets pop it off and put it into our decode
			 * buffer...  And when the decode buffer is full, write another
			 * line...
			 */
			while (sp > stack)
			{
				*bufptr++ = *(--sp);

				if (--bufcnt == 0)
				{
					if ((ret = out_line(buf, linewidth, interlaced, BitDepth)) < 0)
					{
						DeleteObj(buf);
						return(ret);
					}

					bufptr = buf;
					bufcnt = linewidth;
				}
			}
		}
	}

	ret = 0;

	if (bufcnt != linewidth)
	{
		ret = out_line(buf, (linewidth - bufcnt), interlaced, BitDepth);
	}

	DeleteObj(buf);

	return(ret);
}

bool GdcGif::ReadImage(GdcBasePrimitives *pdc)
{
	bool Status = FALSE;
	pDC = pdc;

	if (pDC AND Open())
	{
		bad_code_count = 0;

		if (NOT FindHeader(0, "GIF8?a", *this))
		{
			// not a gif file
		}
		else
		{
			uchar ImgCode;
			uchar interlace = FALSE;

			// read header
			Seek(4, SEEK_CUR);
			*this >> ImgCode;
			int bits = (ImgCode & 0x07) + 1;
			Seek(2, SEEK_CUR);

			if (ImgCode & 0x80)
			{
				uchar PalData[768];

				int size = (1 << bits) * 3;
				memset(PalData, 0xFF, sizeof(PalData));
				Read(PalData, size);

				GdcPalette *Pal = NEW(GdcPalette(PalData));
				if (Pal)
				{
					pDC->Palette(Pal);
				}
			}

			// Start reading the block stream
			uchar BlockCode = 0;
			bool Done = FALSE;

			while (NOT Done AND NOT Eof())
			{
				*this >> BlockCode;

				switch (BlockCode)
				{
					case 0x21:
					{
						Seek(7, SEEK_CUR);
						break;
					}

					case 0x2C:
					{
						Seek(4, SEEK_CUR);

						short sx;
						*this >> sx;
						short sy;
						*this >> sy;
						linewidth = (sx * bits + 7) / 8;

						*this >> interlace;
						interlace = (interlace & 0x40) != 0;

						if (pDC->Create(sx, sy, 8))
						{
							decoder(bits, interlace);
							Status = TRUE;
						}

						Done = TRUE;
						break;
					}
				}
			}
		}
	}

	return Status;
}

bool GdcGif::WriteImage(GdcBasePrimitives *pDC)
{
	return FALSE;
}

